@@ -315,7 +315,85 @@ class SeqLinear(nn.Module):
         
         if self.W_0 is not None: nn.init.constant_(self.W_0, 0.)
     
-    def forward(self, x, W_0=None):
+    def bidirectional(self, x):
+        bsz, src_len, d_model = x.size()
+        
+        x = self.in_proj(x)
+        
+        x, dt = torch.split(
+            x,
+            [
+                2 * self.u_dim + self.v_dim,
+                self.n_heads,
+            ],
+            dim=-1
+        )
+        
+        if cuda_causal_conv1d is not None:
+            x_f = cuda_causal_conv1d(
+                x.transpose(1, 2).contiguous(),
+                self.conv.weight.squeeze(1),
+                bias=self.conv.bias,
+                activation="silu",
+            ).transpose(1, 2)
+            x_b = cuda_causal_conv1d(
+                x.flip(1).transpose(1, 2).contiguous(),
+                self.conv.weight.squeeze(1),
+                bias=self.conv.bias,
+                activation="silu",
+            ).transpose(1, 2)
+        else:
+            x_f = F.silu(self.conv(x.transpose(1, 2).contiguous()).transpose(1, 2)[:, :src_len])
+            x_b = F.silu(self.conv(x.flip(1).transpose(1, 2).contiguous()).transpose(1, 2)[:, :src_len])
+        
+        z_f, u_f, v_f = torch.split(
+            x_f,
+            [
+                self.u_dim,
+                self.u_dim,
+                self.v_dim,
+            ],
+            dim=-1
+        )
+        z_b, u_b, v_b = torch.split(
+            x_b,
+            [
+                self.u_dim,
+                self.u_dim,
+                self.v_dim,
+            ],
+            dim=-1
+        )
+        
+        t = torch.exp(torch.linspace(0, 1, src_len, device=self.device).unsqueeze(-1) * self.dt_rate)
+        dt_f = F.softplus(dt + self.dt_bias) * t
+        z_f = rearrange(z_f, 'b s (h d) -> b s h d', h=self.n_heads).contiguous()
+        u_f = rearrange(u_f, 'b s (h d) -> b s h d', h=self.n_heads).contiguous()
+        v_f = rearrange(v_f, 'b s (h d) -> b s h d', h=self.n_heads).contiguous()
+        
+        dt_b = F.softplus(dt + self.dt_bias).flip(1) * t
+        z_b = rearrange(z_b, 'b s (h d) -> b s h d', h=self.n_heads).contiguous()
+        u_b = rearrange(u_b, 'b s (h d) -> b s h d', h=self.n_heads).contiguous()
+        v_b = rearrange(v_b, 'b s (h d) -> b s h d', h=self.n_heads).contiguous()
+        
+        W_f = torch.cumsum(torch.einsum('bsh, bshd, bshe -> bshde', dt, u_f, v_f), dim=1)
+        W_b = torch.cumsum(torch.einsum('bsh, bshd, bshe -> bshde', dt_b, u_b, v_b), dim=1)
+        
+        if self.W_0 is not None:
+            W_f = W_f + self.W_0.unsqueeze(1)
+            W_b = W_b + self.W_0.unsqueeze(1)
+        
+        norm_f = torch.cumsum(dt_f, dim=1).unsqueeze(-1)
+        norm_b = torch.cumsum(dt_b, dim=1).unsqueeze(-1)
+        
+        out_f = torch.matmul(z_f.unsqueeze(-2), W_f).squeeze(-2) / norm_f
+        out_b = torch.matmul(z_b.unsqueeze(-2), W_b).squeeze(-2) / norm_b
+        out = rearrange(out_f + out_b, 'b s h d -> b s (h d)')
+        out = self.out_proj(out)
+        return out
+    
+    def forward(self, x, mode="causal"):
+        if mode == "bidirectional": return self.bidirectional(x)
         bsz, src_len, d_model = x.size()
         
         x = self.in_proj(x)
@@ -355,9 +433,12 @@ class SeqLinear(nn.Module):
         u = rearrange(u, 'b s (h d) -> b s h d', h=self.n_heads).contiguous()
         v = rearrange(v, 'b s (h d) -> b s h d', h=self.n_heads).contiguous()
         
-        W = torch.cumsum(torch.einsum('bsh, bshd, bshe -> bshde', dt, u, v), dim=1)
-        if W_0 is not None: W = W + W_0.unsqueeze(1)
-        elif self.W_0 is not None: W = W + self.W_0.unsqueeze(1)
+        if mode == "causal":
+            W = torch.cumsum(torch.einsum('bsh, bshd, bshe -> bshde', dt, u, v), dim=1)
+        elif mode == "noncausal":
+            W = torch.einsum('bsh, bshd, bshe -> bhde', dt, u, v).unsqueeze(1)
+        
+        if self.W_0 is not None: W = W + self.W_0.unsqueeze(1)
         
         norm = torch.cumsum(dt, dim=1).unsqueeze(-1)
         
@@ -372,15 +453,14 @@ class SeqMLP(nn.Module):
                  u_dim=None, v_dim=None, W_0=False,
                  d_conv=4, pos_encoding=False, pos_encoding_max_len=None,
                  use_embedding=True, weight_tying=False,
-                 bidirectional=True,
-                 bias=True, device="cpu"):
+                 mode="causal", bias=True, device="cpu"):
         super().__init__()
         self.emb_dim = emb_dim
         self.input_dim = input_dim
         self.output_dim = output_dim
         self.n_layers = n_layers
         self.n_heads = n_heads
-        self.bidirectional = bidirectional
+        self.mode = mode
         self.use_embedding = use_embedding
         self.device = device
 
@@ -393,7 +473,7 @@ class SeqMLP(nn.Module):
         if pos_encoding:
             assert pos_encoding_max_len is not None, "pos_encoding_max_len must be provided for absolute positional encoding"
             self.pos_encoding_max_len = pos_encoding_max_len
-            self.abs_pos_encoding = nn.Embedding(pos_encoding_max_len, emb_dim, device=device) if pos_encoding == "abs" else None
+            self.abs_pos_encoding = nn.Embedding(pos_encoding_max_len, emb_dim, device=device)
         
         self.layers = nn.ModuleList([
             nn.ModuleDict(
@@ -431,9 +511,8 @@ class SeqMLP(nn.Module):
             x = x + self.abs_pos_encoding(pos)
         for layer in self.layers:
             x_norm = layer.norm(x)
-            y_f = layer.dropout(layer.mixer(x_norm))
-            y_b = layer.dropout(layer.mixer(x_norm.flip(1))) if self.bidirectional else 0.0
-            x = x + y_f + y_b
+            y = layer.dropout(layer.mixer(x_norm, mode=self.mode))
+            x = x + y
         x = self.norm_f(x)
         x = self.out_proj(x)
         return x
\ No newline at end of file
