@@ -40,7 +40,7 @@ class Transformer(nn.Module):
         elif pos_encoding == "abs":
             assert pos_encoding_max_len is not None, "pos_encoding_max_len must be provided for absolute positional encoding"
             self.pos_encoding_max_len = pos_encoding_max_len
-            self.abs_pos_encoding = nn.Embedding(pos_encoding_max_len, emb_dim, device=device) if pos_encoding == "abs" else None
+            self.abs_pos_encoding = nn.Embedding(pos_encoding_max_len, emb_dim, device=device)
         
         self.layers = nn.ModuleList([
             nn.ModuleDict(
@@ -118,7 +118,7 @@ class LinearTransformer(nn.Module):
         elif pos_encoding == "abs":
             assert pos_encoding_max_len is not None, "pos_encoding_max_len must be provided for absolute positional encoding"
             self.pos_encoding_max_len = pos_encoding_max_len
-            self.abs_pos_encoding = nn.Embedding(pos_encoding_max_len, emb_dim, device=device) if pos_encoding == "abs" else None
+            self.abs_pos_encoding = nn.Embedding(pos_encoding_max_len, emb_dim, device=device)
         
         self.layers = nn.ModuleList([
             nn.ModuleDict(
@@ -195,7 +195,7 @@ class OrthoLinearTransformer(nn.Module):
         elif pos_encoding == "abs":
             assert pos_encoding_max_len is not None, "pos_encoding_max_len must be provided for absolute positional encoding"
             self.pos_encoding_max_len = pos_encoding_max_len
-            self.abs_pos_encoding = nn.Embedding(pos_encoding_max_len, emb_dim, device=device) if pos_encoding == "abs" else None
+            self.abs_pos_encoding = nn.Embedding(pos_encoding_max_len, emb_dim, device=device)
         
         self.layers = nn.ModuleList([
             nn.ModuleDict(
@@ -274,7 +274,7 @@ class CompressionTransformer(nn.Module):
         elif pos_encoding == "abs":
             assert pos_encoding_max_len is not None, "pos_encoding_max_len must be provided for absolute positional encoding"
             self.pos_encoding_max_len = pos_encoding_max_len
-            self.abs_pos_encoding = nn.Embedding(pos_encoding_max_len, emb_dim, device=device) if pos_encoding == "abs" else None
+            self.abs_pos_encoding = nn.Embedding(pos_encoding_max_len, emb_dim, device=device)
         
         self.layers = nn.ModuleList([
             nn.ModuleDict(
@@ -355,7 +355,7 @@ class SlidingWindowTransformer(nn.Module):
         elif pos_encoding == "abs":
             assert pos_encoding_max_len is not None, "pos_encoding_max_len must be provided for absolute positional encoding"
             self.pos_encoding_max_len = pos_encoding_max_len
-            self.abs_pos_encoding = nn.Embedding(pos_encoding_max_len, emb_dim, device=device) if pos_encoding == "abs" else None
+            self.abs_pos_encoding = nn.Embedding(pos_encoding_max_len, emb_dim, device=device)
         
         dilation_factor = window_len if dilation_factor is None else dilation_factor
         self.layers = nn.ModuleList([
